#include <cppcore/conversion/time.h>

#include <cppamqp/helper.h>
#include <cppamqp/message.inl>
#include <cppamqp/exception.inl>
#include <cppamqp/connection.inl>
#include <cppamqp/consume_result.inl>
#include <cppamqp/publish_options.inl>

using namespace ::cppamqp;

connection::internal::internal() :
    connection   (amqp_new_connection()),
    reply_code   (AMQP_REPLY_SUCCESS)
{
    if (!connection)
        throw cppamqp::exception("unable to create connection");
}

connection::internal::~internal()
{
    if (auto_close)
    {
        auto_close = false;
        __impl::check_and_raise(
            amqp_connection_close(connection, reply_code),
            "error while closing connection",
            true);
    }
    __impl::check_and_raise(
        amqp_destroy_connection(connection),
        "error while destroying connection",
        true);
}


struct internal_envelope : public amqp_envelope_t
{
    internal_envelope()
        { memset(this, 0, sizeof(*this)); }

    ~internal_envelope()
        { amqp_destroy_envelope(this); }
};

struct internal_message : public amqp_message_t
{
    internal_message()
        { memset(this, 0, sizeof(*this)); }

    ~internal_message()
        { amqp_destroy_message(this); }
};

void connection::tcp_connect(const std::string& hostname, uint port)
{
    _internal.reset(new internal());

    auto socket = amqp_tcp_socket_new(handle());
    if (!socket)
        throw exception(std::string("unable to open tcp socket to [") + hostname + "]:" + std::to_string(port));

    __impl::check_and_raise(
        amqp_socket_open(socket, hostname.c_str(), static_cast<int>(port)),
        "error while opening socket",
        false);

    _internal->auto_close = true;
}

void connection::login_plain(
    const std::string&  p_username,
    const std::string&  p_password,
    const std::string&  p_vhost,
    int                 p_heartbeat,
    int                 p_max_channels,
    int                 p_max_frame_size)
{
    __impl::check_and_raise(
        amqp_login(handle(), p_vhost.c_str(), p_max_channels, p_max_frame_size, p_heartbeat, AMQP_SASL_METHOD_PLAIN, p_username.c_str(), p_password.c_str()),
        "error while login plain",
        false);
}

void connection::login_external(
    const std::string&  p_identify,
    const std::string&  p_vhost,
    int                 p_heartbeat,
    int                 p_max_channels,
    int                 p_max_frame_size)
{
    __impl::check_and_raise(
        amqp_login(handle(), p_vhost.c_str(), p_max_channels, p_max_frame_size, p_heartbeat, AMQP_SASL_METHOD_EXTERNAL, p_identify.c_str()),
        "error while login external",
        false);
}

channel connection::open_channel(channel_number c)
    { return channel(*this, c); }

consume_result connection::consume_message(const std::chrono::milliseconds& timeout)
{
    ::timeval tv = ::cppcore::duration_cast<::timeval>(timeout);
    internal_envelope envelope;
    amqp_maybe_release_buffers(handle());
    auto ret = amqp_consume_message(
                    handle(),
                    &envelope,
                    timeout.count() >= 0 ? &tv : nullptr,
                    0);

    if (ret.reply_type == AMQP_RESPONSE_NORMAL)
        return consume_result(*this, envelope);

    if (    ret.reply_type    == AMQP_RESPONSE_LIBRARY_EXCEPTION
        &&  ret.library_error == AMQP_STATUS_TIMEOUT) // TODO: test this case!!!!
        return consume_result(*this, consume_result_type::timeout);

    if (    ret.reply_type    != AMQP_RESPONSE_LIBRARY_EXCEPTION
        ||  ret.library_error != AMQP_STATUS_UNEXPECTED_STATE)
        __impl::check_and_raise(ret, "error while consuming message", false);

    amqp_frame_t frame;
    memset(&tv, 0, sizeof(tv));
    __impl::check_and_raise(
        amqp_simple_wait_frame_noblock(handle(), &frame, &tv),
        "error wile waiting for frame",
        false);
    if (frame.frame_type != AMQP_FRAME_METHOD)
        throw exception("unexpected or unknown frame type");
    switch(frame.payload.method.id)
    {
        case AMQP_BASIC_ACK_METHOD:
            return consume_result(*this, consume_result_type::acknowledge);

        case AMQP_BASIC_RETURN_METHOD:
        {
            internal_message message;
            __impl::check_and_raise(
                amqp_read_message(handle(), frame.channel, &message, 0),
                "error while reading returned message",
                false);
            return consume_result(*this, consume_result_type::could_not_deliver_return_to_sender, message);
        }

        case AMQP_CHANNEL_CLOSE_METHOD:
            throw exception("channel is closed");

        case AMQP_CONNECTION_CLOSE_METHOD:
            close(AMQP_REPLY_SUCCESS, true);
            return consume_result(*this, consume_result_type::connection_closed_by_peer);

        default:
            throw exception("received unexpected frame method");
    }
}

void connection::close(int status, bool force)
{
    if (!_internal)
        return;
    if (force)
        _internal->auto_close = false;
    _internal->reply_code = status;
    _internal.reset();
}
